var crypto = require('crypto');
var debug = require('debug')('node-red-contrib-allenandheathilive');
var tcp = require('net');
var eventEmitter = require("events");
var timeoutPeriod = 1000;
class EventEmitter extends eventEmitter{}
var modes = require("./modes.js");

module.exports = function(RED)
{
    //Send Command
    sendCommand: function(msg, server, midiChannel, answerback) {
            var value = false;
            Object.keys(modes).forEach(function(mode){
                    if(value === false) {
                        var mode = modes[mode];
                        value = mode.send(msg, server, midiChannel);
                    }
                });

            if((typeof value === "string")) {
                //An Error Occurred
                answerback("Mode Error: " + value); 
            }
            else if(value != false){
                answerback(true);
            }
    }

    //Add a listener
    addListener: function(iLiveNetwork, type, fn) {
        iLiveNetwork.event.on(type, fn);
    }

    //Main functiion
    function iLiveNetwork(config)
    {
        //const emitter = new EventEmitter();
        RED.nodes.createNode(this, config);
        console.log("Created iLive Network Node @ IP: " + config.ipAddress + ":" + config.port);
        var midiChannel = config.midiChannel - 1;
        var ipAddress = config.ipAddress;
        var port = config.port;
        this.server = new tcp.Socket();
        this.event = new EventEmitter();
        var node = this;
        var connected = true;
        var connectedWatchDog;
        var disconnectedWatchDog;

        //On close disconnect
        this.on("close", function() {
            server.destroy();
            clearInterval(disconnectedWatchDog);
            clearInterval(connectedWatchDog);
        });

        //Connect
        var connectSuccessful = function() {
            clearInterval(disconnectedWatchDog);
            connected = true;
            this.event.emit("status", "[YELLOW]Connected, No Response");
        
            //Ping the console every 5 seconds to make sure we're still connected to it
            connectedWatchDog = setInterval(function(){
                pingConsole(server, midiChannel, 

                //Success
                function() {
                    connected = true;
                    this.event.emit("status", "[GREEN]Connected!");
                }
                ,
                //Failed
                function() {
                    connected = false;
                    this.event.emit("status", "[RED]Disconnected");
                });
            } , 5000);
        }
        server.connect(port, ipAddress, connectSuccessful);

        //Socket Closed - Try to reconnect
        server.on("close", function() {
            if(connected) {  
                disconnectedWatchDog = setInterval(function() {
                    server.connect(port, ipAddress);
                }, 5000);      
                connected = false;
                server.once("connect", connectSuccessful);
                clearInterval(connectedWatchDog);
            }
        });

        //Error
        server.on("error", function(error){
            this.event.emit("status", "[RED]Disconnected");
            server.destroy();
        });

        //Got data in 
        server.on("data", function(message){
            connected = true;

            var value = false;
            Object.keys(modes).forEach(function(mode){
                    if(value === false) {
                        var mode = modes[mode];
                        value = mode.recieve(midiChannel, message);
                    }
                });

            if((typeof value === "string")) {
                //An Error Occurred
                RED.log.error("Mode Error: " + value);
                this.event.emit("status", "[RED]An Error Occured");
            }
            else if(value != false){
                this.event.emit("status", "[GREEN]Got Request");
                node.send(value);
            }
        });
    }

    RED.nodes.registerType("allenandheathilive-ilivenetwork", iLiveNetwork);
}

//Ping the console to make sure it's still connected
function pingConsole(server, midiChannel, successFunction, failedFunction)
{
    var success = false;

    //Get a channel colour, just to get a reply
    var buffer = new Buffer(12);
    buffer.writeUInt8(0xF0, 0);
    buffer.writeUInt8(0x00, 1);
    buffer.writeUInt8(0x00, 2);
    buffer.writeUInt8(0x1A, 3);
    buffer.writeUInt8(0x50, 4);
    buffer.writeUInt8(0x10, 5);
    buffer.writeUInt8(0x01, 6);
    buffer.writeUInt8(0x00, 7);
    buffer.writeUInt8((0x00 + parseInt(midiChannel, 16)), 8);
    buffer.writeUInt8(0x04, 9);
    buffer.writeUInt8(0x20, 10);
    buffer.writeUInt8(0xF7, 11);
    server.write(buffer);

    var handler = function(message){
        if(message[12] == 0xF7) {success=true;}
        successFunction();
    }
    server.once("data", handler);

    //Timeout check
    setTimeout(function() {
        if(success == false) {
            server.removeListener("data", handler);
            failedFunction();
        }
    }, timeoutPeriod);
}
